to268 updated this revision to Diff 537331.
to268 added a comment.

I have fixed the misleading diagnostic with auto array declarations.
The diagnostics now says that `'auto' is not allowed in array declarations`.

I am not sure with the way i have currently implemented it in ` SemaDecl.cpp`, 
maybe it can be done in a cleaner way.


Repository:
  rG LLVM Github Monorepo

CHANGES SINCE LAST ACTION
  https://reviews.llvm.org/D133289/new/

https://reviews.llvm.org/D133289

Files:
  clang/include/clang/Basic/DiagnosticSemaKinds.td
  clang/lib/Parse/ParseDecl.cpp
  clang/lib/Sema/DeclSpec.cpp
  clang/lib/Sema/SemaDecl.cpp
  clang/lib/Sema/SemaTemplateDeduction.cpp
  clang/test/C/C2x/n3007.c
  clang/test/CodeGen/auto.c
  clang/test/Parser/c2x-auto.c
  clang/test/Sema/c2x-auto.c
  clang/www/c_status.html

Index: clang/www/c_status.html
===================================================================
--- clang/www/c_status.html
+++ clang/www/c_status.html
@@ -1190,12 +1190,12 @@
     <tr>
       <td>Underspecified object definitions</td>
       <td><a href="https://www.open-std.org/jtc1/sc22/wg14/www/docs/n3006.htm";>N3006</a></td>
-      <td class="unknown" align="center">Unknown</td>
+      <td class="none" align="center">Unknown</td>
     </tr>
     <tr>
       <td>Type inference for object declarations</td>
       <td><a href="https://www.open-std.org/jtc1/sc22/wg14/www/docs/n3007.htm";>N3007</a></td>
-      <td class="none" align="center">No</td>
+      <td class="unreleased" align="center">Clang 17</td>
     </tr>
     <tr>
       <td>constexpr for object definitions</td>
Index: clang/test/Sema/c2x-auto.c
===================================================================
--- /dev/null
+++ clang/test/Sema/c2x-auto.c
@@ -0,0 +1,119 @@
+// RUN: %clang_cc1 -std=c2x -verify -pedantic -Wno-comments %s
+
+void test_basic_types(void) {
+  auto undefined;                 // expected-error {{declaration of variable 'undefined' with deduced type 'auto' requires an initializer}}
+  auto auto_int = 4;
+  auto auto_long = 4UL;
+
+  _Static_assert(_Generic(auto_int, int : 1));
+  _Static_assert(_Generic(auto_long, unsigned long : 1));
+}
+
+void test_sizeof_typeof(void) {
+  auto auto_size = sizeof(auto);  // expected-error {{expected expression}}
+  typeof(auto) tpof = 4;          // expected-error {{expected expression}}
+}
+
+void test_casts(void) {
+  auto int_cast = (int)(4 + 3);
+  auto double_cast = (double)(1 / 3);
+  auto long_cast = (long)(4UL + 3UL);
+  auto auto_cast = (auto)(4 + 3); // expected-error {{expected expression}}
+
+  _Static_assert(_Generic(int_cast, int : 1));
+  _Static_assert(_Generic(double_cast, double : 1));
+  _Static_assert(_Generic(long_cast, long : 1));
+}
+
+void test_compound_literral(void) {
+  auto int_cl = (int){13};
+  auto double_cl = (double){2.5};
+  auto array[] = { 1, 2, 3 }; // expected-error {{cannot use 'auto' with initializer list in C}}
+
+  auto auto_cl = (auto){13};  // expected-error {{expected expression}}
+
+  _Static_assert(_Generic(int_cl, int : 1));
+  _Static_assert(_Generic(double_cl, double : 1));
+}
+
+void test_array_pointers(void) {
+  double array[3] = { 0 };
+  auto a = array;
+  auto b = &array;
+
+  _Static_assert(_Generic(array, double * : 1));
+  _Static_assert(_Generic(a, double * : 1));
+  _Static_assert(_Generic(b, double (*)[3] : 1));
+}
+
+void test_typeof() {
+  int typeof_target();
+  auto result = (typeof(typeof_target())){12};
+
+  _Static_assert(_Generic(result, int : 1));
+}
+
+void test_qualifiers(const int y) {
+  const auto a = 12;
+  auto b = y;
+  static auto c = 1UL;
+  int* pa = &a; // expected-warning {{initializing 'int *' with an expression of type 'const int *' discards qualifiers}}
+  const int* pb = &b;
+  int* pc = &c; // expected-warning {{incompatible pointer types initializing 'int *' with an expression of type 'unsigned long *'}}
+
+  _Static_assert(_Generic(a, int : 1));
+  _Static_assert(_Generic(b, int : 1));
+  _Static_assert(_Generic(c, unsigned long : 1));
+  _Static_assert(_Generic(pa, int * : 1));
+  _Static_assert(_Generic(pb, const int * : 1));
+  _Static_assert(_Generic(pc, int * : 1));
+}
+
+void test_strings(void) {
+  auto str = "this is a string";
+  auto str2[] = "this is a string";
+  auto (str3) = "this is a string";
+  auto (((str4))) = "this is a string";
+
+  _Static_assert(_Generic(str, char * : 1));
+  _Static_assert(_Generic(str2, char * : 1));
+  _Static_assert(_Generic(str3, char * : 1));
+  _Static_assert(_Generic(str4, char * : 1));
+}
+
+void test_pointers(void) {
+  auto a = 12;
+  auto *ptr = &a;                         // expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  auto *str = "this is a string";         // expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  const auto *str2 = "this is a string";  // expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  auto *b = &a;                           // expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  *b = &a;                                // expected-error {{incompatible pointer to integer conversion assigning to 'int' from 'int *'; remove &}}
+  auto nptr = nullptr;
+
+  _Static_assert(_Generic(a, int : 1));
+  _Static_assert(_Generic(ptr, int * : 1));
+  _Static_assert(_Generic(str, char * : 1));
+  _Static_assert(_Generic(str2, const char * : 1));
+  _Static_assert(_Generic(b, int * : 1));
+}
+
+void test_scopes(void) {
+  double a = 7;
+  double b = 9;
+  {
+    auto a = a * a; // expected-error {{variable 'a' declared with deduced type 'auto' cannot appear in its own initializer}} \
+                       expected-error {{variable 'a' declared with deduced type 'auto' cannot appear in its own initializer}}
+  }
+  {
+    auto b = a * a;
+    auto a = b;
+
+    _Static_assert(_Generic(b, double : 1));
+    _Static_assert(_Generic(a, double : 1));
+  }
+}
+
+[[clang::overloadable]] auto test(auto x) { // expected-error {{'auto' not allowed in function prototype}} \
+                                               expected-error {{'auto' not allowed in function return type}}
+  return x;
+}
Index: clang/test/Parser/c2x-auto.c
===================================================================
--- /dev/null
+++ clang/test/Parser/c2x-auto.c
@@ -0,0 +1,124 @@
+// RUN: %clang_cc1 -fsyntax-only -verify=expected,c2x -std=c2x %s
+// RUN: %clang_cc1 -fsyntax-only -verify=expected,c17 -std=c17 %s
+
+#define AUTO_MACRO(_NAME, ARG, ARG2, ARG3) \
+auto _NAME = ARG + (ARG2 / ARG3);
+
+struct S {
+  int a;
+  auto b;       // c2x-error {{'auto' not allowed in struct member}} \
+                   c17-error {{type name does not allow storage class to be specified}} \
+                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+  union {
+    char c;
+    auto smth;  // c2x-error {{'auto' not allowed in union member}} \
+                   c17-error {{type name does not allow storage class to be specified}} \
+                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+  } u;
+};
+
+enum E : auto { // c2x-error {{'auto' not allowed here}} \
+                   c17-error {{expected a type}} \
+                   c17-error {{type name does not allow storage class to be specified}}
+  One,
+  Two,
+  Tree,
+};
+
+auto basic_usage(auto auto) {   // c2x-error {{'auto' not allowed in function prototype}} \
+                                   c2x-error {{'auto' not allowed in function return type}} \
+                                   c2x-error {{cannot combine with previous 'auto' declaration specifier}} \
+                                   c17-error {{invalid storage class specifier in function declarator}} \
+                                   c17-error {{illegal storage class on function}} \
+                                   c17-warning {{duplicate 'auto' declaration specifier}} \
+                                   c17-warning {{omitting the parameter name in a function definition is a C2x extension}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto = 4;                     // expected-error {{expected identifier or '('}}
+
+  auto a = 4;                   // c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto auto aa = 12;            // c2x-error {{cannot combine with previous 'auto' declaration specifier}} \
+                                   c17-warning {{duplicate 'auto' declaration specifier}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto b[4];                    // c2x-error {{'auto' not allowed in array declaration}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto array[auto];             // expected-error {{expected expression}} \
+                                   c2x-error {{declaration of variable 'array' with deduced type 'auto' requires an initializer}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  AUTO_MACRO(auto, 1, 2, 3);    // c2x-error {{cannot combine with previous 'auto' declaration specifier}} \
+                                   expected-error {{expected identifier or '('}} \
+                                   c17-warning {{duplicate 'auto' declaration specifier}}
+
+  auto c = (auto)a;             // expected-error {{expected expression}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto ci = (auto){12};         // expected-error {{expected expression}} \
+                                   c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  return c;
+}
+
+void structs(void) {
+  struct s_auto { auto a; };            // c2x-error {{'auto' not allowed in struct member}} \
+                                           c17-error {{type name does not allow storage class to be specified}} \
+                                           c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto s_int = (struct { int a; } *)0;  // c17-error {{incompatible pointer to integer conversion initializing 'int' with an expression of type}} \
+                                           c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  typedef auto auto_type;               // c2x-error {{'auto' not allowed in typedef}} \
+                                           c17-error {{cannot combine with previous 'typedef' declaration specifier}} \
+                                           c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+}
+
+void sizeof_alignas(void) {
+  auto auto_size = sizeof(auto);  // expected-error {{expected expression}} \
+                                     c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+}
+
+void generic_alignof_alignas(void) {
+  int g;
+  _Generic(g, auto : 0);  // c2x-error {{'auto' not allowed here}} \
+                             c17-error {{expected a type}} \
+                             c17-error {{type name does not allow storage class to be specified}}
+
+  _Alignof(auto);         // expected-error {{expected expression}} \
+                             expected-warning {{'_Alignof' applied to an expression is a GNU extension}}
+
+  _Alignas(auto);         // expected-error {{expected expression}} \
+                             expected-warning {{declaration does not declare anything}}
+}
+
+void test_function_pointers(void) {
+  extern auto auto_ret_func(void);    // c2x-error {{'auto' not allowed in function return type}} \
+                                         c17-error {{cannot combine with previous 'extern' declaration specifier}} \
+                                         c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  extern void auto_param_func(auto);  // c2x-error {{'auto' not allowed in function prototype}} \
+                                         c17-error {{invalid storage class specifier in function declarator}} \
+                                         c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  auto (auto_ret_func)(void);         // c2x-error {{'auto' not allowed in function return type}} \
+                                         c17-error {{illegal storage class on function}} \
+                                         c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  void (auto_param_func)(auto);       // c2x-error {{'auto' not allowed in function prototype}} \
+                                         c17-error {{invalid storage class specifier in function declarator}} \
+                                         c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+}
+
+void atomic(void) {
+  _Atomic(auto) atom1 = 12; // c2x-error {{'auto' not allowed here}} \
+                               c2x-error {{a type specifier is required for all declarations}} \
+                               c17-error {{expected a type}} \
+                               c17-error {{type name does not allow storage class to be specified}} \
+                               c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+
+  _Atomic auto atom2 = 12;  // c2x-error {{_Atomic cannot be applied to type 'auto' which is not trivially copyable}} \
+                               c17-error {{type specifier missing, defaults to 'int'; ISO C99 and later do not support implicit int}}
+}
Index: clang/test/CodeGen/auto.c
===================================================================
--- /dev/null
+++ clang/test/CodeGen/auto.c
@@ -0,0 +1,39 @@
+// RUN: %clang_cc1 -std=c2x -emit-llvm %s -o - | FileCheck %s
+
+void basic_types(void) {
+  auto nb = 4;          // CHECK: %nb = alloca i32, align 4
+  auto dbl = 4.3;       // CHECK: %dbl = alloca double, align 8
+  auto lng = 4UL;       // CHECK: %lng = alloca i{{32|64}}, align {{4|8}}
+  auto bl = true;       // CHECK: %bl = alloca i8, align 1
+  auto chr = 'A';       // CHECK: %chr = alloca i32, align 4
+  auto str = "Test";    // CHECK: %str = alloca ptr, align 8
+  auto str2[] = "Test"; // CHECK: %str2 = alloca [5 x i8], align 1
+  auto nptr = nullptr;  // CHECK: %nptr = alloca ptr, align 8
+}
+
+void misc_declarations(void) {
+  auto strct_ptr = (struct { int a; } *)0;  // CHECK: %strct_ptr = alloca ptr, align 8
+  auto int_cl = (int){13};                  // CHECK: %int_cl = alloca i32, align 4
+  auto double_cl = (double){2.5};           // CHECK: %double_cl = alloca double, align 8
+
+  auto se = ({      // CHECK: %se = alloca i32, align 4
+    auto snb = 12;  // CHECK: %snb = alloca i32, align 4
+    snb;
+  });
+}
+
+void loop(void) {
+  auto j = 4;                       // CHECK: %j = alloca i32, align 4
+  for (auto i = j; i < 2 * j; i++); // CHECK: %i = alloca i32, align 4
+}
+
+#define AUTO_MACRO(_NAME, ARG, ARG2, ARG3) auto _NAME = ARG + (ARG2 / ARG3);
+
+#define AUTO_INT_MACRO(_NAME, ARG, ARG2, ARG3) auto _NAME = (ARG ^ ARG2) & ARG3;
+
+int macros(int in_int) {
+  auto a = in_int + 1;             // CHECK: %a = alloca i32, align 4
+  AUTO_MACRO(b, 1.3, 2.5f, 3);     // CHECK: %b = alloca double, align 8
+  AUTO_INT_MACRO(c, 64, 23, 0xff); // CHECK: %c = alloca i32, align 4
+  return (a + (int)b) - c;         // CHECK: ret i32 %sub
+}
Index: clang/test/C/C2x/n3007.c
===================================================================
--- /dev/null
+++ clang/test/C/C2x/n3007.c
@@ -0,0 +1,182 @@
+// RUN: %clang_cc1 -std=c2x -verify -pedantic -Wno-comments %s
+
+/* WG14 N3007: Yes
+ * Type Inference for object definitions
+ */
+void test_qualifiers(int x, const int y) {
+  const auto a = x;
+  auto b = y;
+  static auto c = 1UL;
+  int* pa = &a; // expected-warning {{initializing 'int *' with an expression of type 'const int *' discards qualifiers}}
+  const int* pb = &b;
+  int* pc = &c; // expected-warning {{incompatible pointer types initializing 'int *' with an expression of type 'unsigned long *'}}
+
+  const int ci = 12;
+  auto yup = ci;
+  yup = 12;
+
+  _Static_assert(_Generic(a, int : 1));
+  _Static_assert(_Generic(c, unsigned long : 1));
+  _Static_assert(_Generic(pa, int * : 1));
+  _Static_assert(_Generic(pb, const int * : 1));
+}
+
+void test_atomic(void) {
+  _Atomic auto i = 12;  // expected-error {{_Atomic cannot be applied to type 'auto' which is not trivially copyable}}
+  _Atomic(auto) j = 12; // expected-error {{'auto' not allowed here}} \
+                           expected-error {{a type specifier is required for all declarations}}
+
+  _Atomic(int) foo(void);
+  auto k = foo();
+
+  _Static_assert(_Generic(&i, _Atomic auto *: 1)); // expected-error {{_Atomic cannot be applied to type 'auto' which is not trivially copyable}} \
+                                                      expected-error {{'auto' not allowed here}}
+  _Static_assert(_Generic(k, int: 1));
+}
+
+void test_double(void) {
+  double A[3] = { 0 };
+  auto pA = A;
+  auto qA = &A;
+  auto pi = 3.14;
+
+  _Static_assert(_Generic(A, double * : 1));
+  _Static_assert(_Generic(pA, double * : 1));
+  _Static_assert(_Generic(qA, double (*)[3] : 1));
+  _Static_assert(_Generic(pi, double : 1));
+}
+
+int test_auto_param(auto a) { // expected-error {{'auto' not allowed in function prototype}}
+  return (int)(a * 2);
+}
+
+auto test_auto_return(float a, int b) { // expected-error {{'auto' not allowed in function return type}}
+  return ((a * b) * (a / b));
+}
+
+[[clang::overloadable]] auto test(auto x) { // expected-error {{'auto' not allowed in function prototype}} \
+                                               expected-error {{'auto' not allowed in function return type}}
+  return x;
+}
+
+void test_sizeof_alignas(void) {
+  auto auto_size = sizeof(auto);  // expected-error {{expected expression}}
+  _Alignas(auto) int b[4];        // expected-error {{expected expression}}
+}
+
+void test_arrary(void) {
+  auto a[4];          // expected-error {{'auto' not allowed in array declaration}}
+  auto b[] = {1, 2};  // expected-error {{cannot use 'auto' with initializer list in C}}
+}
+
+void test_initializer_list(void) {
+  auto a = {};        // expected-error {{cannot use 'auto' with initializer list in C}}
+  auto b = { 0 };     // expected-error {{cannot use 'auto' with initializer list in C}}
+  auto c = { 1, };    // expected-error {{cannot use 'auto' with initializer list in C}}
+  auto d = { 1 , 2 }; // expected-error {{cannot use 'auto' with initializer list in C}}
+  auto e = (int [3]){ 1, 2, 3 };
+}
+
+void test_structs(void) {
+  // FIXME: This should be diagnosed as invalid as described in N3006
+  auto p1 = (struct { int a; } *)0;
+  struct s;
+  auto p2 = (struct s { int a; } *)0;
+
+  struct B { auto b; };   // expected-error {{'auto' not allowed in struct member}}
+}
+
+void test_typedefs(void) {
+  typedef auto auto_type;   // expected-error {{'auto' not allowed in typedef}}
+
+  typedef auto (*fp)(void); // expected-error {{'auto' not allowed in typedef}}
+  typedef void (*fp)(auto); // expected-error {{'auto' not allowed in function prototype}}
+
+  _Generic(0, auto : 1);    // expected-error {{'auto' not allowed here}}
+}
+
+void test_misc(void) {
+  auto something;                 // expected-error {{declaration of variable 'something' with deduced type 'auto' requires an initializer}}
+  auto test_char = 'A';
+  auto test_char_ptr = "test";
+  auto test_char_ptr2[] = "another test";
+  auto auto_size = sizeof(auto);  // expected-error {{expected expression}}
+
+  _Static_assert(_Generic(test_char, int : 1));
+  _Static_assert(_Generic(test_char_ptr, char * : 1));
+  _Static_assert(_Generic(test_char_ptr2, char * : 1));
+}
+
+void test_no_integer_promotions(void) {
+  short s;
+  auto a = s;
+  _Generic(a, int : 1); // expected-error {{controlling expression type 'short' not compatible with any generic association type}}
+}
+
+void test_compound_literral(void) {
+  auto a = (int){};
+  auto b = (int){ 0 };
+  auto c = (int){ 0, };
+  auto d = (int){ 0, 1 };       // expected-warning {{excess elements in scalar initializer}}
+
+  auto auto_cl = (auto){13};    // expected-error {{expected expression}}
+
+  _Static_assert(_Generic(a, int : 1));
+  _Static_assert(_Generic(b, int : 1));
+  _Static_assert(_Generic(c, int : 1));
+}
+
+void test_pointers(void) {
+  int a;
+  auto *ptr = &a; // expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  auto *ptr2 = a; // expected-error {{variable 'ptr2' with type 'auto *' has incompatible initializer of type 'int'}} \
+                     expected-warning {{explicit declaration of 'auto' pointers is a Clang extension}}
+  auto nptr = nullptr;
+
+  _Static_assert(_Generic(ptr, int * : 1));
+  _Static_assert(_Generic(ptr2, int * : 1));
+}
+
+void test_scopes(void) {
+  double a = 7;
+  double b = 9;
+  {
+    auto a = a * a; // expected-error {{variable 'a' declared with deduced type 'auto' cannot appear in its own initializer}} \
+                       expected-error {{variable 'a' declared with deduced type 'auto' cannot appear in its own initializer}}
+  }
+  {
+    auto b = a * a;
+    auto a = b;
+
+    _Static_assert(_Generic(a, double : 1));
+    _Static_assert(_Generic(b, double : 1));
+  }
+}
+
+void test_loop(void) {
+  auto j = 4;
+  for (auto i = j; i < 2 * j; i++);
+
+  _Static_assert(_Generic(j, int : 1));
+}
+
+#define AUTO_MACRO(_NAME, ARG, ARG2, ARG3) \
+auto _NAME = ARG + (ARG2 / ARG3);
+
+// This macro should only work with integers due to the usage of binary operators
+#define AUTO_INT_MACRO(_NAME, ARG, ARG2, ARG3) \
+auto _NAME = (ARG ^ ARG2) & ARG3;
+
+void test_macros(int in_int) {
+  auto a = in_int + 1;
+  AUTO_MACRO(b, 1.3, 2.5f, 3);
+  AUTO_INT_MACRO(c, 64, 23, 0xff);
+  AUTO_INT_MACRO(not_valid, 51.5, 25, 0xff); // expected-error {{invalid operands to binary expression ('double' and 'int')}}
+
+  auto result = (a + (int)b) - c;
+
+  _Static_assert(_Generic(a, int : 1));
+  _Static_assert(_Generic(b, double : 1));
+  _Static_assert(_Generic(c, int : 1));
+  _Static_assert(_Generic(result, int : 1));
+}
Index: clang/lib/Sema/SemaTemplateDeduction.cpp
===================================================================
--- clang/lib/Sema/SemaTemplateDeduction.cpp
+++ clang/lib/Sema/SemaTemplateDeduction.cpp
@@ -4777,9 +4777,24 @@
     return TDK_Success;
   }
 
+  // Make sure that we treat 'char[]' equaly as 'char*' in C2x mode
+  auto *String = dyn_cast<StringLiteral>(Init);
+  if (getLangOpts().C2x && String && Type.getType()->isArrayType()) {
+    TypeLoc TL = TypeLoc(Init->getType(), Type.getOpaqueData());
+    Result = SubstituteDeducedTypeTransform(*this, DependentResult).Apply(TL);
+    assert(!Result.isNull() && "substituting DependentTy can't fail");
+    return TDK_Success;
+  }
+
+  // Emit a warning if 'auto*' is used in pedantic and in C2x mode
+  if (getLangOpts().C2x && Type.getType()->isPointerType()) {
+    Diag(Type.getBeginLoc(), diag::ext_c2x_auto_pointer_declaration);
+  }
+
   auto *InitList = dyn_cast<InitListExpr>(Init);
   if (!getLangOpts().CPlusPlus && InitList) {
-    Diag(Init->getBeginLoc(), diag::err_auto_init_list_from_c);
+    Diag(Init->getBeginLoc(), diag::err_auto_init_list_from_c)
+        << (AT->isGNUAutoType() ? "__auto_type" : "auto");
     return TDK_AlreadyDiagnosed;
   }
 
Index: clang/lib/Sema/SemaDecl.cpp
===================================================================
--- clang/lib/Sema/SemaDecl.cpp
+++ clang/lib/Sema/SemaDecl.cpp
@@ -12681,6 +12681,17 @@
   DeducedType *Deduced = Type->getContainedDeducedType();
   assert(Deduced && "deduceVarTypeFromInitializer for non-deduced type");
 
+  // Diagnose auto array declarations in C2x,
+  // unless it's a char array or an initialiser list which is diagnosed later
+  if (getLangOpts().C2x && Type->isArrayType() &&
+      !isa_and_nonnull<StringLiteral>(Init) &&
+      !isa_and_nonnull<InitListExpr>(Init)) {
+    Diag(Range.getBegin(), diag::err_auto_not_allowed)
+      << (int)Deduced->getContainedAutoType()->getKeyword()
+      << 23 << Range;
+    return QualType();
+  }
+
   // C++11 [dcl.spec.auto]p3
   if (!Init) {
     assert(VDecl && "no init for init capture deduction?");
Index: clang/lib/Parse/ParseDecl.cpp
===================================================================
--- clang/lib/Parse/ParseDecl.cpp
+++ clang/lib/Parse/ParseDecl.cpp
@@ -3960,7 +3960,7 @@
       isStorageClass = true;
       break;
     case tok::kw_auto:
-      if (getLangOpts().CPlusPlus11) {
+      if (getLangOpts().CPlusPlus11 || getLangOpts().C2x) {
         if (isKnownToBeTypeSpecifier(GetLookAheadToken(1))) {
           isInvalid = DS.SetStorageClassSpec(Actions, DeclSpec::SCS_auto, Loc,
                                              PrevSpec, DiagID, Policy);
Index: clang/include/clang/Basic/DiagnosticSemaKinds.td
===================================================================
--- clang/include/clang/Basic/DiagnosticSemaKinds.td
+++ clang/include/clang/Basic/DiagnosticSemaKinds.td
@@ -237,6 +237,9 @@
   "imaginary constants are a GNU extension">, InGroup<GNUImaginaryConstant>;
 def ext_integer_complex : Extension<
   "complex integer types are a GNU extension">, InGroup<GNUComplexInteger>;
+def ext_c2x_auto_pointer_declaration : Extension<
+  "explicit declaration of 'auto' pointers is a Clang extension">,
+  InGroup<DiagGroup<"auto-decl-extensions">>;
 
 def err_invalid_saturation_spec : Error<"'_Sat' specifier is only valid on "
   "'_Fract' or '_Accum', not '%0'">;
@@ -2337,7 +2340,8 @@
   "|in conversion function type|here|in lambda parameter"
   "|in type allocated by 'new'|in K&R-style function parameter"
   "|in template parameter|in friend declaration|in function prototype that is "
-  "not a function declaration|in requires expression parameter}1">;
+  "not a function declaration|in requires expression parameter"
+  "|in array declaration}1">;
 def err_dependent_deduced_tst : Error<
   "typename specifier refers to "
   "%select{class template|function template|variable template|alias template|"
@@ -2410,7 +2414,7 @@
 def err_malformed_std_initializer_list : Error<
   "std::initializer_list must be a class template with a single type parameter">;
 def err_auto_init_list_from_c : Error<
-  "cannot use __auto_type with initializer list in C">;
+  "cannot use '%0' with initializer list in C">;
 def err_auto_bitfield : Error<
   "cannot pass bit-field as __auto_type initializer in C">;
 
_______________________________________________
cfe-commits mailing list
cfe-commits@lists.llvm.org
https://lists.llvm.org/cgi-bin/mailman/listinfo/cfe-commits

Reply via email to
Index: clang/lib/Sema/DeclSpec.cpp
===================================================================
--- clang/lib/Sema/DeclSpec.cpp
+++ clang/lib/Sema/DeclSpec.cpp
@@ -1363,8 +1363,9 @@
     StorageClassSpecLoc = SourceLocation();
   }
   // Diagnose if we've recovered from an ill-formed 'auto' storage class
-  // specifier in a pre-C++11 dialect of C++.
-  if (!S.getLangOpts().CPlusPlus11 && TypeSpecType == TST_auto)
+  // specifier in a pre-C++11 dialect of C++ or in a C2X dialect of C.
+  if ((!S.getLangOpts().CPlusPlus11 && !S.getLangOpts().C2x) &&
+      TypeSpecType == TST_auto)
     S.Diag(TSTLoc, diag::ext_auto_type_specifier);
   if (S.getLangOpts().CPlusPlus && !S.getLangOpts().CPlusPlus11 &&
       StorageClassSpec == SCS_auto)